import imagej
import scyjava
import scyjava

import os


class BaseImageJHelper:

    def __init__(self, headless=False, version="2.1.0", plugins= [], plugins_dir="plugins") -> None:
        # concatenade version plus plugins
        build = [version] + plugins if len(plugins) > 0 else version
        if plugins_dir:
            path = os.path.join(os.getcwd(),plugins_dir)
            scyjava.config.add_option(f'-Dplugins.dir={path}')

        self.headless = headless
        self._ij = imagej.init(version, headless=headless)
        if not headless: self._ij.ui().showUI()
        super().__init__()
    
    @property
    def py(self):
        return self._ij.py

    @property
    def ui(self):
        return self._ij.ui()

    @property
    def ij(self):
        return self._ij

RUNNING_HELPER = None

def get_running_helper() -> BaseImageJHelper:
    global RUNNING_HELPER
    if RUNNING_HELPER is None:
        RUNNING_HELPER = BaseImageJHelper()
    return RUNNING_HELPER

def set_running_helper(instance: BaseImageJHelper):
    global RUNNING_HELPER
    RUNNING_HELPER = instance


from mikro.schema import Representation
from mikroj.registries.helper import BaseImageJHelper, set_running_helper
import dask

class ImageJHelper(BaseImageJHelper):

    def __init__(self, headless=False, bind=True, version="sc.fiji:fiji:2.1.0", plugins = []) -> None:
        if bind: set_running_helper(self)
        super().__init__(headless=headless, version=version, plugins= plugins)


    def displayRep(self, rep: Representation):
        image = rep.data.squeeze()

        if dask.is_dask_collection(image.data):
            jimage = self.py.to_java(image.compute())
        else:
            jimage = self.py.to_java(image)

        # Convert the Image to Image
        self.ui.show(rep.name, jimage)


from mikroj.macros.filter import FilterMacro


color_code = FilterMacro("""
                                var Glut = "Fire";	//default LUT
                    var Gstartf = 1;
                    var Gendf = 10;
                    var GFrameColorScaleCheck = 1;
                    var GbatchMode = 0;

                    macro "Time-Lapse Color Coder" {
                        Stack.getDimensions(ww, hh, channels, slices, frames);
                        if (channels > 1)
                            exit("Cannot color-code multi-channel images!");
                        //swap slices and frames in case:
                        if ((slices > 1) && (frames == 1)) {
                            frames = slices;
                            slices = 1;
                            Stack.setDimensions(1, slices, frames);
                            print("slices and frames swapped");
                        }
                        Gendf = frames;
                        if (Gstartf <1) Gstartf = 1;
                        if (Gendf > frames) Gendf = frames;
                        totalframes = Gendf - Gstartf + 1;
                        calcslices = slices * totalframes;
                        imgID = getImageID();

                        setBatchMode(true);

                        newImage("colored", "RGB White", ww, hh, calcslices);
                        run("Stack to Hyperstack...", "order=xyczt(default) channels=1 slices="
                            + slices + " frames=" + totalframes + " display=Color");
                        newimgID = getImageID();

                        selectImage(imgID);
                        run("Duplicate...", "duplicate");
                        run("8-bit");
                        imgID = getImageID();

                        newImage("stamp", "8-bit White", 10, 10, 1);
                        run(Glut);
                        getLut(rA, gA, bA);
                        close();
                        nrA = newArray(256);
                        ngA = newArray(256);
                        nbA = newArray(256);

                        newImage("temp", "8-bit White", ww, hh, 1);
                        tempID = getImageID();

                        for (i = 0; i < totalframes; i++) {
                            colorscale = floor((256 / totalframes) * i);
                            for (j = 0; j < 256; j++) {
                                intensityfactor = j / 255;
                                nrA[j] = round(rA[colorscale] * intensityfactor);
                                ngA[j] = round(gA[colorscale] * intensityfactor);
                                nbA[j] = round(bA[colorscale] * intensityfactor);
                            }

                            for (j = 0; j < slices; j++) {
                                selectImage(imgID);
                                Stack.setPosition(1, j + 1, i + Gstartf);
                                run("Select All");
                                run("Copy");

                                selectImage(tempID);
                                run("Paste");
                                setLut(nrA, ngA, nbA);
                                run("RGB Color");
                                run("Select All");
                                run("Copy");
                                run("8-bit");

                                selectImage(newimgID);
                                Stack.setPosition(1, j + 1, i + 1);
                                run("Select All");
                                run("Paste");
                            }
                        }

                        selectImage(tempID);
                        close();

                        selectImage(imgID);
                        close();

                        selectImage(newimgID);

                        run("Stack to Hyperstack...", "order=xyctz channels=1 slices="
                            + totalframes + " frames=" + slices + " display=Color");
                        op = "start=1 stop=" + Gendf + " projection=[Max Intensity] all";
                        run("Z Project...", op);
                        if (slices > 1)
                            run("Stack to Hyperstack...", "order=xyczt(default) channels=1 slices=" + slices
                                + " frames=1 display=Color");
                        resultImageID = getImageID();

                        selectImage(newimgID);
                        close();

                        selectImage(resultImageID);
                        
                        if (GbatchMode == 0)
                            setBatchMode("exit and display");
                        }
            """)



import re
from ..registries.helper import get_running_helper

argument_matcher = re.compile(r'[^\#]*\#\@\s*(?P<type>[^\s]*)\s*(?P<name>[^\s]*)')

class BaseMacro:

    def __init__(self, macro: str) -> None:
        self.macro = macro
        self.required_kwargs = {}

        # Lets check the for arguments
        for line in iter(self.macro.splitlines()):
            m = argument_matcher.match(line)
            if m:
                self.required_kwargs[m.group("name")] = m.group("type")

        super().__init__()

    def run(self,  with_result=True, helper = None, **kwargs):
        for key, value in self.required_kwargs.items():
            assert key in kwargs, f"This macro needs these arguments: {[f'{key}: {value}' for key, value in self.required_kwargs.items()]}"

        myhelper = helper or get_running_helper()
        result = myhelper.py.run_macro(self.macro, kwargs)
        if with_result: return result



from ..registries.helper import get_running_helper
from ..macros.base import BaseMacro
import xarray as xr
import dask

class ImageActingMacro(BaseMacro):

    def run(self, image: xr.DataArray, helper= None, **kwargs) -> xr.DataArray:
        helper = helper or get_running_helper()
        image = image.squeeze()

        # check if we are having a dask array
        if dask.is_dask_collection(image.data):
            jimage = helper.py.to_java(image.compute())
        else:
            jimage = helper.py.to_java(image)
        # Convert the Image to Image
        if not helper.headless: helper.ui.show(image.name, jimage)
        #run the Macro
        result = super().run(**kwargs)
        print(result)
        # open synchronized image
        imp = helper.py.active_image_plus()
        array = helper.py.from_java(imp)
        if "Channel" in array.dims: array = array.rename({"Channel": "c"})
        return array



from typing import Any
from ..macros.image_acting import ImageActingMacro
import xarray as xr
import dask

class FilterMacro(ImageActingMacro):

    def __call__(self, image: xr.DataArray, **kwargs: Any) -> xr.DataArray:
        return super().run(image, **kwargs)


from arkitekt.schema.widgets import SearchWidget
import numpy
from mikroj.helper import ImageJHelper
from mikro.schema import Representation, RepresentationVariety
from arkitekt import AppAgent, agents, use, QueryWidget
import argparse
import xarray as xr
import numpy as np
from mikroj.implementations.temp_color_code import color_code




def main(**agent_params):

    agent = AppAgent(**agent_params)

    helper = ImageJHelper()


    @agent.register(widgets={
        "rep": SearchWidget(query="""
            query Search($search: String){
                options: representations(name: $search) {
                    value: id
                    label: name
                }
            }
        """)
    })
    def show(rep: Representation) -> Representation:
        """Show

        Shows the Image on MikroJ

        Args:
            rep (Representation): [description]

        Returns:
            Representation: [description]
        """
        helper.displayRep(rep)
        return rep

    @agent.register()
    def color_coder(rep: Representation) -> Representation:
        """Color Code Z

        Runs the Image Macro Color Code 

        Args:
            rep (Representation): The Image Coming in

        Returns:
            Representation: The Color Coded Image
        """
        image = color_code(rep.data)
        return Representation.objects.from_xarray(image, sample=rep.sample, meta=rep.meta, variety=RepresentationVariety.VOXEL, tags=["rgb"], name=f"Color Code of {rep.name}")

    agent.provide()




if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', help='Which config file to use', default="bergen.yaml", type=str)
    args = parser.parse_args()

    main(config_path=args.config)








